import java.util.Properties;

/**
 * Title: Einstellungen
 * Description: Einstellungen save the options made in TheHive
 * @author Daniel Schlosser & Dimitri Missoh
 * @version 1.0
 */

class Einstellungen extends Parameters {

    //Deklaration von Variablen
    private Simulator theSim;
    private TheHive theHive;
    private ModellViewer viewport;

    private int    standartRundenzahl=500;
    private double mutationProb=0.1;
    private int    genpoolSize=100;
    private int conflictResolvingSt=theSim.FIRST;
    private int roundsWithoutUser=10;
    private boolean automaticFitness=false;
    private boolean interactionMode=false;
    private int sizeOfUmgebungen=71;
    private int numberAgents=5;
    private int numberGensInGenom=10;
    private int sizeOfView=3;
    private boolean doCrossover=true;
    private int positionOfCrystallizationPoint=0;

    private String standartRundenzahlName   = "standartRundenzahl";
    private String mutationProbName         = "mutationProb";
    private String genpoolSizeName          = "genpoolSize";
    private String numberToSurviveName      = "numberToSurvive";
    private String conflictResolvingStName  = "conflictResolvingSt";
    private String roundsWithoutUserName    = "roundsWithoutUser";
    private String automaticFitnessName     = "automaticFitness";
    private String interactionModeName      = "interactionMode";
    private String sizeOfUmgebungenName     = "sizeOfUmgebungen";
    private String numberAgentsName         = "numberAgents";
    private String numberGensInGenomName    = "numberGensInGenom";
    private String sizeOfViewName           = "sizeOfView";
    private String doCrossoverName           = "doCrossover";
    private String positionOfCrystallizationPointName="positionOfCrystallizationPoint";
   /**
    *Konstruktor der Klasse Einstelungen.
    *Hier werden die Default-Parameter gelesen und gesetzt,
    *und dann gespeichert
    */
   Einstellungen(Simulator theSim, TheHive theHive,ModellViewer viewport) {
   	super("swarm.ser", "Swarm-Build Properties");
        this.theSim=theSim;
        this.theHive=theHive;
	this.viewport=viewport;
   	getParameters();
   	saveParameters();
        theHive.setStandartRundenZahl(this.standartRundenzahl);
        theSim.setPropability(this.mutationProb);
        theSim.setNumberNests(this.genpoolSize);
        theSim.setResolveOfConflict(this.conflictResolvingSt);
        theHive.setRoundsWithoutUser(this.roundsWithoutUser);
        theHive.setAutomaticFitness(this.automaticFitness);
        theHive.setInteractionMode(this.interactionMode);
        theSim.setSizeOfNest(new Koord3D(this.sizeOfUmgebungen,this.sizeOfUmgebungen,this.sizeOfUmgebungen));
        theSim.setNumberAgents(this.numberAgents);
        theSim.setNumberGensInGenom(this.numberGensInGenom);
	theSim.setDoCrossover(this.doCrossover);
	theSim.setPositionCrystallizationPoint(this.positionOfCrystallizationPoint);
	//viewport.setSizeOfNest(())
    }

    /**
     *diese Methode dient dazu, die Variablen auf den Default-Werten einzusetzen.
     *param defaults vom Typ Properties
     *return hat keinen Rckgabewert
     */
    protected void setDefaults(Properties defaults) {

        defaults.put(standartRundenzahlName, new Integer(standartRundenzahl).toString());
        defaults.put(mutationProbName, new Double(mutationProb).toString());
        defaults.put(genpoolSizeName, new Integer(genpoolSize).toString());
        defaults.put(conflictResolvingStName, new Integer(conflictResolvingSt).toString());
	defaults.put(roundsWithoutUserName, new Integer(roundsWithoutUser).toString());
	defaults.put(automaticFitnessName, new Boolean(automaticFitness).toString());
	defaults.put(interactionModeName, new Boolean(interactionMode).toString());
	defaults.put(sizeOfUmgebungenName, new Integer(sizeOfUmgebungen).toString());
	defaults.put(numberAgentsName, new Integer(numberAgents).toString());
	defaults.put(numberGensInGenomName, new Integer(numberGensInGenom).toString());
	defaults.put(sizeOfViewName,new Integer(this.sizeOfView).toString()) ;
	defaults.put(doCrossoverName,new Boolean(this.doCrossover).toString());
	defaults.put(this.positionOfCrystallizationPointName,new Integer(this.positionOfCrystallizationPoint).toString());
    }

    /**
     *diese Methode dient dazu, die Variablen auf den gespeicherten Werten einzusetzen.
     *param - erwartet keinen Paramater
     *return hat keinen Rckgabewert
     */
    protected void updateSettingsFromProperties() {
        try {
            standartRundenzahl  = Integer.parseInt(properties.getProperty(standartRundenzahlName));
            mutationProb        = Double.parseDouble(properties.getProperty(mutationProbName));
            genpoolSize         = Integer.parseInt(properties.getProperty(genpoolSizeName));
            conflictResolvingSt = Integer.parseInt(properties.getProperty(conflictResolvingStName));
	    roundsWithoutUser   = Integer.parseInt(properties.getProperty(roundsWithoutUserName));
	    automaticFitness    = (Boolean.valueOf(properties.getProperty(automaticFitnessName))).booleanValue();
	    interactionMode     = (Boolean.valueOf(properties.getProperty(interactionModeName))).booleanValue();
	    sizeOfUmgebungen    = Integer.parseInt(properties.getProperty(sizeOfUmgebungenName));
	    numberAgents        = Integer.parseInt(properties.getProperty(numberAgentsName));
	    numberGensInGenom   = Integer.parseInt(properties.getProperty(numberGensInGenomName));
	    sizeOfView          = Integer.parseInt(properties.getProperty(sizeOfViewName));
	    doCrossover         = (Boolean.valueOf(properties.getProperty(doCrossoverName))).booleanValue();
	    positionOfCrystallizationPoint=Integer.parseInt(properties.getProperty(positionOfCrystallizationPointName));
        } catch (NumberFormatException e) {
			//wir kmmern uns nicht darum wenn das property im falschen Format ist,
			//sie werden alle die default-Werte annehmen.
        }
    }

    /**
     *diese Methode dient dazu, die gespeicherten Werten auf den aktuellen Werten einzusetzen.
     *param - erwartet keinen Paramater
     *return hat keinen Rckgabewert
     */
    protected void updatePropertiesFromSettings() {
	properties.put(standartRundenzahlName, new Integer(standartRundenzahl).toString());
	properties.put(mutationProbName, new Double(mutationProb).toString());
	properties.put(genpoolSizeName, new Integer(genpoolSize).toString());
        properties.put(conflictResolvingStName, new Integer(conflictResolvingSt).toString());
	properties.put(roundsWithoutUserName, new Integer(roundsWithoutUser).toString());
	properties.put(automaticFitnessName,new Boolean(automaticFitness).toString());
	properties.put(interactionModeName, new Boolean(interactionMode).toString());
	properties.put(sizeOfUmgebungenName, new Integer(sizeOfUmgebungen).toString());
	properties.put(numberAgentsName,new Integer(numberAgents).toString());
	properties.put(numberGensInGenomName, new Integer(numberGensInGenom).toString());
	properties.put(sizeOfViewName,new Integer(sizeOfView).toString());
	properties.put(doCrossoverName,new Boolean(doCrossover).toString());
	properties.put(this.positionOfCrystallizationPointName,new Integer(this.positionOfCrystallizationPoint).toString());
    }

   /**
     *diese Methode dient dazu, die gespeicherten Werten in einem String umzuwandeln.
     *param - erwartet keinen Paramater
     *return hat keinen Rckgabewert
     */
    public String toString() {
        return "[\n"
               + "standartRundenzahl=" + standartRundenzahl + ",\n"
               + "mutationProb="       + mutationProb + ",\n"
               + "genpoolSize="        + genpoolSize + ",\n"
               + "conflictResolvingSt" + conflictResolvingSt + ",\n"
	       + "roundsWithoutUser" + roundsWithoutUser + ",\n"
	       + "automaticFitness" + automaticFitness + ",\n"
	       + "interactionMode" + interactionMode + ",\n"
	       + "sizeOfUmgebungen" + sizeOfUmgebungen + ",\n"
	       + "numberAgents" + numberAgents + ",\n"
	       + "numberGensInGenom" + numberGensInGenom + ",\n"
	       + "sizeOfView" + this.sizeOfView + ",\n"
	       + "doCrossover" + this.doCrossover + ",\n"
	       + "positionOfCrystalizationPoint" + this.positionOfCrystallizationPoint
               +"]";
	}

    //die darunter liegenden Methoden dienen dazu,
    //die gespeicherte Werte einzusetzen
    //und einzulesen
    void setSizeOfView(int sizeOfView){
	this.sizeOfView=sizeOfView;
	saveParameters();
	theSim.setSizeOfView(sizeOfView);
    }
    int getSizeOfView(){
	return this.sizeOfView;
    }
    void setDoCrossover(boolean doCrossover){
	this.doCrossover=doCrossover;
	saveParameters();
	theSim.setDoCrossover(doCrossover);
    }
    boolean getDoCrossover(){
	return this.doCrossover;
    }
    void setPositionOfCrystalizationPoint(int i){
	this.positionOfCrystallizationPoint=i;
	saveParameters();
	theSim.setPositionCrystallizationPoint(i);
    }
    int getPositionOfCrystalizationPoint(){
	return this.positionOfCrystallizationPoint;
    }
    void setStandartRundenzahl(int standartRundenzahl) {
        this.standartRundenzahl = standartRundenzahl;
        saveParameters();
        theHive.setStandartRundenZahl(standartRundenzahl);
    }

    int getStandartRundenzahl() {
        return standartRundenzahl;
    }

    void setMutationProb(double mutationProb) {
        this.mutationProb = mutationProb;
        saveParameters();
        theSim.setPropability(mutationProb);
    }

    double getMutationProb() {
    	return mutationProb;
    }

    void setGenpoolSize(int genpoolSize) {
      this.genpoolSize = genpoolSize;
      saveParameters();
      theSim.setNumberNests(genpoolSize);
    }

    int getGenpoolSize() {
      return genpoolSize;
    }

    void setConflictResolvingSt(int conflictResolvingSt) {
      this.conflictResolvingSt = conflictResolvingSt;
      saveParameters();
      theSim.setResolveOfConflict(conflictResolvingSt);
    }

    int getConflictResolvingSt() {
      return conflictResolvingSt;
    }

    void setRoundsWithoutUser(int roundsWithoutUser) {
      this.roundsWithoutUser = roundsWithoutUser;
      saveParameters();
      theHive.setRoundsWithoutUser(roundsWithoutUser);
    }

    int getRoundsWithoutUser() {
        return roundsWithoutUser;
    }

    void setAutomaticFitness(boolean automaticFitness) {
      this.automaticFitness = automaticFitness;
      saveParameters();
      theHive.setAutomaticFitness(automaticFitness);
    }

    boolean getAutomaticFitness() {
      return automaticFitness;
    }

    void setInteractionMode(boolean interactionMode) {
      this.interactionMode = interactionMode;
      saveParameters();
      theHive.setInteractionMode(interactionMode);
    }

    boolean getInteractionMode() {
      return interactionMode;
    }

    void setSizeOfUmgebungen(int sizeOfUmgebungen) {
      this.sizeOfUmgebungen = sizeOfUmgebungen;
      saveParameters();
      theSim.setSizeOfNest(new Koord3D(sizeOfUmgebungen,sizeOfUmgebungen,sizeOfUmgebungen));
    }

    int getSizeOfUmgebungen() {
        return sizeOfUmgebungen;
    }

    void setNumberAgents(int numberAgents) {
      this.numberAgents = numberAgents;
      saveParameters();
      theSim.setNumberAgents(numberAgents);
      theSim.resetTotal();
    }

    int getNumberAgents() {
      return numberAgents;
    }

    void setNumberGensInGenom(int numberGensInGenom) {
      this.numberGensInGenom = numberGensInGenom;
      saveParameters();
      theSim.setNumberGensInGenom(numberGensInGenom);
    }

    int getNumberGensInGenom() {
      return numberGensInGenom;
    }

}


