import java.util.LinkedList;
import java.io.*;

/**
 * Die Klasse Environment implementiert die Simulator-Schnittstelle ber die die GUI auf die Funktionalitt des Agenten und der Gene zugreift.
 *
 * Stand 09.07.2001
 * @author Clemens Mhlberger, Jan Bannert, Steffen Glckselig
 * @see simulator
 */

public class Environment
    implements Simulator{

    protected Nest currentNest;

    protected int resolveConflictByStrategy;
    protected double mutationPropability;

    protected int numberNests;
    protected int numberAgents;
    protected static int numberGens;  // Anzahl der Gene in einem Genom, wird in Environment(int) verwendet
    protected GenPool Pool;
    protected int indexCurrentNest = -1;

    protected Koord3D sizeNest;

    protected int posCrystallizationPoint;
    protected static final int STANDARD_CRYSTALL_POINT = 0;

    protected long builtCounter = 0;    // 14.07
    protected boolean enableCrossover;  // 14.07
    protected static final boolean STANDARD_DO_CROSSOVER = true;

    protected static final int STANDARD_NGinG    = 5;     // 04.07.2001 (Steffen)
    protected static final int STANDARD_NAgents  = 4;
    protected static final int STANDARD_sizeSide = 71;

    protected static final int STANDARD_viewSize = 3;
    protected int viewSize;

    protected static final int STANDARD_NESTs = 100;

    protected final double STANDARD_PropStein    = 1.5;
    protected final double STANDARD_PropJoker    = 0;
    protected final double STANDARD_PropLeer     = 27;

    protected static final int STANDARD_RESOLVE  = Simulator.LEAST_JOKER;


    /**
     * Der Standardkonstruktor fr eine Environment. Es wird die LinkedList, die zur Haltung der Nesten verwendet wird, und ein GenPool initialisiert.
     *
     * Gendert: 02.07.2001 -> sizeNest + Number_Gens_...
     *           17.07.2001 ruft jetzt den in der Hierarchie nextn Konstruktor auf
     */
    public Environment(){
	this( STANDARD_NESTs );
    }

    /**
     * Ein Konstruktor fr eine Environment.
     * Neben der LinkedList wird zustzlich die angegebene Anzahl von Nesten initialisiert und zur Liste hinzugefgt.
     * <br>Zustzlich wird als standardmssig als Konfliktlsung LEAST_JOKER gewhlt, also das Muster mit den wenigsten Wildcards.
     *
     * Gendert: 02.07.2001 -> sizeNest + Number_Gens_...
     */
    public Environment( int number ){
        this( number, STANDARD_NAgents );
    }

    /**
     * Zustzlich noch mit Anzahl der Agenten in einer Nest.
     */
    public Environment( int numbNests, int numbAgents ){
        this( numbNests, numbAgents, new Koord3D(STANDARD_sizeSide,STANDARD_sizeSide,STANDARD_sizeSide));
    }

    /**
     * Zustzlich mit Grsse der Nest.
     */
    public Environment( int numbNests, int numbAgents, Koord3D size ){
        this( numbNests, numbAgents, size, STANDARD_NGinG);
    }

    /**
     * Zustzlich mit Anzahl der Gene pro Genom.
     */
    public Environment( int numbNests, int numbAgents, Koord3D size, int numbGens ){
        this( numbNests, numbAgents, size, numbGens, STANDARD_RESOLVE );
    }

    /**
     * Zustzlich mit Grsse des Sichtfelds.
     */
    public Environment( int numbNests, int numbAgents, Koord3D size, int numbGens, int strategy ){
        this( numbNests, numbAgents, size, numbGens, strategy, STANDARD_viewSize );
    }


    public Environment( int numbNests, int numbAgents, Koord3D size, int numbGens, int strategy, int viewSize ){
        this( numbNests, numbAgents, size, numbGens, strategy, viewSize, STANDARD_DO_CROSSOVER );
    }

    /**
     * Zustzlich mit Konfliktauflsungsstrategie.
     */
    public Environment( int numbNests, int numbAgents, Koord3D size, int numbGens, int strategy, int viewSize, boolean enableCrossover ){
	this( numbNests, numbAgents, size, numbGens, strategy, viewSize, enableCrossover, STANDARD_CRYSTALL_POINT );
    }

    public Environment( int numbNests, int numbAgents, Koord3D size, int numbGens, int strategy, int viewSize, boolean enableCrossover, int startingPoint ){
        sizeNest = size;
	posCrystallizationPoint = startingPoint;
        this.viewSize = viewSize;
        this.enableCrossover = enableCrossover;
        numberGens = numbGens;

        numberNests  = numbNests;
        numberAgents = numbAgents;

        Pool = new GenPool( numberNests, numberGens, viewSize, STANDARD_PropStein, STANDARD_PropJoker, STANDARD_PropLeer );
        resolveConflictByStrategy = strategy;

        nextNest();  // ++++++++++++++++++++++++
    }

    /** Hiermit kann man die Standardgrsse einer Nest verndern.
     *
     * Gendert: 02.07.2001
     * @param size neue Grsse
     */
    public void setSizeOfNest( Koord3D size ){
        sizeNest = size;
    }

    // 14.07.2001
    public long getNumberBuiltBlocks(){
        return builtCounter;
    }

    public int getNumberOfNests(){
        return Pool.size();
    }

    public int getIndexOfCurrentNest(){
        return indexCurrentNest;
    }

    public void setDoCrossover( boolean abler ){
        enableCrossover = abler;
    }

    public int getHowOftenUsed_Gen(int index ){
        return currentNest.getHowOftenUsed_Gen( index );
    }

    public int getSizeGenom(){
        return currentNest.getGenom().size();
    }

    // 17.0.7.2001
    public void setSizeOfView( int size ){
	viewSize = size;
    }

    public void setPositionCrystallizationPoint( int pos ){
	posCrystallizationPoint = pos;
    }

    /** Hiermit kann man die Anzahl der Gene in einem Genom festlegen.
     *
     * Gendert: 09.07.2001
     * @param number neue Anzahl der Gene
     */
    public void setNumberGensInGenom( int number ){
        numberGens = number;
    }

    /** Hiermit kann man die Anzahl der Agenten, die sich in einer Nest tummeln sollen, festlegen.
     *
     * Gendert: 09.07.2001
     * @param number neue Anzahl von Agenten
     */
    public void setNumberAgents( int number ){
        numberAgents = number;
    }

    /** Liefert die Anzahl der Agenten, die sich in der aktuellen Nest befinden.
     *
     * Gendert: 09.07.2001
     * @return Anzahl der Agenten, die sich gegenwrtig wirklich in der Nest tummeln.
     */
    public int getNumberAgents(){
        return currentNest.getNumberAgents();
    }

    /**
     * Methode setzt die Konfliktlsungsstrategie fr die aktuelle und alle folgenden Nesten.
     *
     * Gendert: 12.06.2001
     *           12.07.2001 null-Pointer abgefangen
     * @param strategy Gibt die Konfliktlsung an
     */
    public void setResolveOfConflict( int strategy ){
        resolveConflictByStrategy = strategy;
        if( currentNest != null ){
            currentNest.setStrategy( strategy );
        }
    }

    /**
     * Methode liefert die Konfliktlsungsstrategie fr die aktuelle Nest.
     *
     * Gendert: 12.06.2001
     * @return Konfliktlsungsstrategie fr diese Nest
     */
    public int getResolveOfConflict(){
        return resolveConflictByStrategy;
    }

    /**
     * Methode lsst den Simulator die angegebene Anzahl von Runden laufen.
     *
     * Gendert: 18.06.2001 komplett auf 3D umgestellt
     *           09.07.2001 Koord3D-Array enthlt jetzt auch Lcken (spter null'en), damit jeder Agent seine eigene Farbe bekommen kann.
     *           14.07.2001 anstelle von (0,0,0) wird in dem Koord3D-Array jetzt null zurckgeliefert  
     * @param runden Anzahl der zu laufenden Runden
     * @return points Koord3D-Array das die Positionen angibt, an denen in den Runden Steine gebaut wurden.
     */
    public Koord3D[] laufeRunden(int runden){
        Koord3D[] result = new Koord3D[runden * numberAgents];

        int index = 0;
        Waypoint[] points = null;

        for( int i = 0; i<runden; i++){
            points = currentNest.runOneRound();
            for( int agent = 0; agent < numberAgents; agent++){
                if( points[agent] != null && points[agent].getBuilt() ){
                    builtCounter++;
                    result[index] = points[agent].getKoord();
                }else{
                    result[index] = null;
                }

                index++;
            }
        }
       
        return result;
    }

    /**
     * Methode liefert alle Bloecke die in der angegebenen Runde gebaut wurden
     *
     * Gendert: 18.06.2001 to3d() entfernt
     *           09.07.2001 Koord3D-Array enthlt jetzt auch Lcken (spter null'en), damit jeder Agent seine eigene Farbe bekommen kann.
     * @param runde Nummer der Runde, von der die Blcke angefordert werden
     * @return array Koord3D-Array, indem die Positionen der gebauten Blcke angegeben werden. <br>Wegen Multiagentensystem -> Array.
     */
    public Koord3D[] gebautInRunde(int runde){
        Koord3D[] result = new Koord3D[numberAgents];
        int counter = 0;
        for( int i =0; i<numberAgents; i++){
            try{
                Waypoint t = currentNest.getAgentsWay(i).getWaypoint( runde );
                if( t.getBuilt() ){
                    result[i] = t.getKoord();
                }else{
                    // hier muss es irgendwann mal auf null gesetzt werden!
                    result[i] = null;
                }

            }catch( IndexOutOfBoundsException e ){
                message( "In Environment.gebautInRunde() wurde eine hoehere Rundenzahl abgefragt als gebaut wurde: "+runde );
            }
        }

        return result;
    }

    /**
     * Methode liefert die Positionen der Agenten(Multiagentensystem) fr die angegebene Runde.
     *
     * Gendert: 18.06.2001 to3d() entfernt
     * @param runde angeforderte Runde
     * @return array mit demutationPropabilityn Positionen der Agenten (zur Zeit DES Agenten)
     */
    public Koord3D[] getAgentPos( int runde ){
        Koord3D[] array = new Koord3D[numberAgents];
        for( int i = 0; i<numberAgents; i++ ){
            try{
                array[i] = currentNest.getAgentsWay(i).getWaypoint( runde ).getKoord();
            }catch( IndexOutOfBoundsException e ){
                message( "In Environment.getAgentPos() wurde eine hoehere Rundenzahl abgefragt als gebaut wurde: "+runde );
            }
        }

        return array;
    }

    /**
     * Liefert Position des ersten Blocks.
     *
     * Gendert: 18.06.2001 to3d() entfernt
     * @return Koord3D mit der Position des ersten Blocks
     */
    public Koord3D getCrystallizationPoint(){
        return currentNest.getCrystallizationPoint();
    }


    /**
     * Wechselt auf nchste Nest in der Liste.
     *
     * Gendert: 02.07.2001 -> auf doMutation() umgestellt, weil makeNewPool() (noch) nicht funktioniert
     *                      -> umgebung.setStrategy() wird jetzt mit der oben gesetzten Variable aufgerufen
     *                      -> neuer Nests-Konstruktor
     *           14.07.2001 makeNewPool() ersetzt mit doMutation() und doCrossover(), wobei doMutation() immer aufgerufen wird.
     * @return boolean ob dieser Wechsel erfolgreich war
     */
    public boolean nextNest(){
        indexCurrentNest++;
        builtCounter = 0;

        // warte alle 10 Lufe extra lang
        if( indexCurrentNest % 10 == 0 ){
            debug( "extra-GC");
            gc(1000);

            // und alle 30 Lufe nochmal so lang
            if( indexCurrentNest % 30 == 0 ){
                gc(2000);
            }
        }

        if( indexCurrentNest > Pool.size() - 1){
            // starte neuen GenPool
            debug("\nErzeuge neuen GenPool durch Crossover/Mutation, lsche alte Nest...");

            finalize();

            Pool = Pool.doMutation( mutationPropability );
            if( enableCrossover ){
                Pool = Pool.doCrossover( numberNests );
            }

            indexCurrentNest = -1;

            nextNest();

            debug( "Environment.nextsNest(): dies ist Nest "+indexCurrentNest+ " von "+Pool.size()+":");

            return false;
        }else{
            finalize();

            debug( "\nEnvironment.nextsNest(): dies ist Nest "+indexCurrentNest+ " von "+Pool.size()+":");

            currentNest = new Nest( Pool.getGenom( indexCurrentNest ), sizeNest, numberAgents, viewSize, posCrystallizationPoint);
            currentNest.setStrategy( resolveConflictByStrategy );

            return true;
        }
    }

    protected void gc(int howLong ){
        System.gc();
        System.gc();
        try{
            Thread.sleep( howLong );
        }catch( InterruptedException e ){
        }
    }

    protected void finalize(){
        if( currentNest != null ){
            currentNest.finalize();
            currentNest = null;
        }

        gc( 500 );
    }

    /** Zum setzen der Mutationswahrscheinlichkeit
     *
     * Gendert: 02.07.2001
     * @param value neue Wahrscheinlichkeit (Wert wird skaliert)
     */
    public void setPropability( double value ){
        mutationPropability = value;
    }

    /** Methode ldt einen kompletten GenPool aus einem Unterverzeichnis (d.h. alle darin enthaltenen txt-Dateien,
     * falls es sich dabei um eine Genom-Datei handelt).
     *
     * Gendert: 05.07.2001
     * @param verzeichnis Unterverzeichnis, in dem sich der GenPool befindet.
     */
    public boolean loadGP( String verzeichnis ){
        File directory = new File( verzeichnis );
        debug( "Environment.loadGP(): " + verzeichnis );

        if( directory.isDirectory() ){
            File[] genFiles = directory.listFiles(new TxtFilenameFilter());
            GenPool temp = new GenPool();

            for( int i = 0; i<genFiles.length; i++ ){
                debug( "Environment.loadGP() laedt Datei: "+genFiles[i].getName());

                Genom genom = loadGenom( genFiles[i].getName() );
                temp.add( genom );
            }

            indexCurrentNest = -1;

            Pool = temp;
            nextNest();
            return true;
        }else{
            message( "Environment.loadGP(): kein Verzeichnis angegeben!");
            return false;
        }
    }


    /**
     * Ldt Nest und Genom. Hierfr werden die .list und die .txt Dateien eingelesen.
     * Aus der .list-Datei wird das Nest rekonstruiert, d.h. die bisher gebauten Steine.
     * Aus der .txt-Datei wird die Position des Agenten, die Grsses der Nest (des Arrays) und das Genom rekonstruiert.
     *
     * Aktuell ersetzt die geladene Nest die aktuelle!!
     *
     * Gendert: 19.06.2001 Nur noch eine Nest, die aktuelle wird dann berschrieben
     *           21.06.     evtl. bergebene Dateierweiterung wird weggeschnitten.
     *           03.07.2001 Es wird nur noch der bergebene Dateiname verwendet, weil es sonst nicht richtig funktioniert
     *           05.07.2001 Geladenes Genom wird in GenPool an die nchste Stelle eingefgt(!)
     * @param verzeichnis Gibt ein Verzeichnis an
     * @param dateiname Gibt eine pseudo-Datei an, an die .list und .txt angehngt wird.
     * @return boolean ob das Laden erfolgreich war oder nicht.
     */
    public boolean load(String verzeichnis, String dateiname){
        String filename = verzeichnis;

        if( new java.io.File( filename ).isFile() ){
            int index = filename.lastIndexOf(".");
            if( index > 0){
                filename = filename.substring(0,index);
            }
        }

        debug("Lade Genom aus Datei "+filename +".gen..." );
        Genom genom = loadGenom( filename + ".gen" );

        if( genom != null ){
            Pool.add( indexCurrentNest + 1, genom);

            return true;
        }else{
            return false;
        }
    }

    protected Genom loadGenom( String filename ){
        try{
            BufferedReader reader = new BufferedReader( new FileReader( filename ) );

            // hier werden die Zeilen, die die Gene enthalten eingelesen.
            // da wir nicht von vornherein wissen, wie viele es gibt, verwenden wir eine dynamische Struktur, einen Vektor.
            String line = reader.readLine();
            if( !line.equals( Nest.HEADER )){
                throw new IOException();
            }

            line = reader.readLine();
            java.util.Vector vector = new java.util.Vector();
            while( line.length() > 0 ){
                vector.add( line );
                line = reader.readLine();
            }
            // es ist etwas umstndlich aus dem Vector ein String[] zu machen, aber es klappt...
            String[] stringArray = new String[vector.size()];
            vector.toArray( stringArray );

            return new Genom( stringArray );
        }catch( IOException e ){
            return null;
        }
    }


    /**
     * Speichert aus der aktuellen Nest die entstandene Struktur, Genom, Position des Agenten und Grsses der Nest.
     *
     * Gendert: 03.07.2001 -> Es wird nur noch der bergebene Dateiname verwendet, weil es sonst nicht richtig funktioniert
     *           05.07.2001 Struktur wird nicht mehr gespeichert
     *
     * @param verzeichnis Gibt ein Verzeichnis an
     * @param dateiname Gibt eine pseudo-Datei an, an die .list und .txt angehngt wird.
     * @return boolean ob das Speichern erfolgreich war oder nicht.
     */
    public boolean store(String verzeichnis,String dateiname){
        String filename = verzeichnis;
        if( new java.io.File( filename ).isFile() ){
            return false;
        }else{
            return store(filename);
        }
    }

    /** Speichert den kompletten GenPool als Genom-txt-Dateien in einem Unterverzeichnis. Diese Funktion berschreibt evtl. bereits vorhandene Dateien!
     *
     * Gendert: 05.07.2001
     * @param verzeichnis Unterverzeichnis, in das der GenPool gespeichert werden soll.
     * @return ob das Speicher funktioniert hat oder nicht.
     */
    public boolean storeGP( String verzeichnis ){
        File directory = new File( verzeichnis );
        directory.mkdir();
        boolean noFailure = true;
        String file = directory.getName()+System.getProperty( "file.separator" )+"Gen_";

        for( int i = 0; i<Pool.size();i++){
            noFailure &= store( "", file+i+".gen" );
        }

        return noFailure;
    }

    /** Mit dieser Funktion kann gespeichert werden selbst wenn die Datei bereits existieren sollte.
     *
     * Gendert: 05.07.2001
     * @param verzeichnis in das zu speichernde Verzeichnis
     * @param dateiname der Name der Datei. Zur Zeit wird nur diese Information verwendet!!
     * @return ob das Speichern geklappt hat.
     */
    public boolean storeAnyways(String verzeichnis,String dateiname){
        String filename = verzeichnis;

        return store(filename);
    }

    /** Speicher ein Genom, berschreibt aber keine bereits bestehende Datei!
     *
     * Gendert: 05.07.2001
     * @param filename Name unter dem die Datei(das Genom) gespeichert werden soll
     * @return ob das Speichern funktioniert hat oder nicht
     */
    protected boolean store( String filename ){
        return currentNest.saveNestAndGenom( filename +".gen" );
    }

    /**
     * Lscht die in der aktuellen Nest entstandene Architektur, aber belsst das Genom
     *
     * Gendert: ?
     * @return boolean ob das Resetten erfolgreich war, oder nicht
     */
    public boolean resetNest(){
	builtCounter = 0;
        currentNest.reset();
        System.gc();System.gc();
	
        return true;
    }

    /**
     * Lscht einfach alle Nesten und erzeugt komplett neue. Deren Anzahl kann mit setNumberNests(int) festgelegt werden.
     * Arbeitet daher ein bisschen anders als resetNest().
     *
     * Gendert: 13.06.2001
     * @return boolean ob das Resetten erfolgreich war, oder nicht
     */
    public boolean resetTotal(){
        Pool = new GenPool( numberNests, numberGens, viewSize, STANDARD_PropStein, STANDARD_PropJoker, STANDARD_PropLeer );

        indexCurrentNest = -1;

        if( currentNest != null ){
            currentNest.finalize();
            currentNest = null;
            System.gc();
        }

        nextNest();

        return true;
    }

    /**
     * Mit dieser Methode kann mit die Anzahl bei resetTotal() zu erzeugenden Nesten festlegen.
     *
     * Gendert: ?
     * @param number Anzahl der zu erzeugenden Nesten
     */
    public void setNumberNests( int number ){
        numberNests = number;
    }


    /** Mit dieser Methode wird die Fitness des aktuellen Genoms gesetzt.
     *
     * Gendert: 09.07.2001 Fitness-Wert wird jetzt an Genom weitergegeben und geht nicht mehr ber GenPool
     *
     * @param value Fitness. Wert wird skaliert.
     */
    public void setFitness( double value ){
        debug( "Environment.setFitness():  Fitness: "+value);
        currentNest.getGenom().setFitness( value );
    }


    // for debugging purpose
    protected static void debug( String msg ){
        System.out.println( msg );
    }


    protected static void message( String msg ){
        System.out.println( msg );
    }
}

