import javax.media.j3d.*;
import com.sun.j3d.utils.geometry.*;
import javax.vecmath.*;
import java.awt.event.MouseEvent;
/**
 * Title: Honeycomb
 * Description: a honeycomb to view in virtual universe
 * @author Dimitri Missoh & Daniel Schlosser
 * @version 1.0
 */

public class Honeycomb extends BranchGroup implements ClickableShape3D{
       private boolean marked,selected;
       private int x,y,z;
       private Transform3D trans;
       private TransformGroup translation;
       private HoneycombShape3D cell;
       private ModellViewer viewport;
    /**
     * Constructor using 3d coordinate and the Modellviewer to setup correctly
     * @param x the x coordinate
     * @param y the y coordinate
     * @param z the z coordinate
     * @param mv the ModellViewer to call if clicked
     */
    public Honeycomb(int x,int y,int z,ModellViewer mv)
    {
           super();
           this.x=x;
           this.y=y;
           this.z=z;
	   viewport=mv;
           marked=false;
           selected=false;
           trans =new Transform3D();
           trans.setTranslation(new Vector3d(x,y,z));
           translation=new TransformGroup(trans);
           cell=new HoneycombShape3D(this);
	   translation.addChild(cell);
           this.setCapability(BranchGroup.ALLOW_DETACH);
           this.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
           this.setCapability(BranchGroup.ENABLE_PICK_REPORTING);
           this.addChild(translation);

    }
    /**
     * You can use this to get information about the position of this honeycomb
     * @return coordinate of this honeycomb
     */
    public Koord3D getpos()
    {
        return new Koord3D(x,y,z);
    }

    /**
     * This will change the color of the honeycomb
     * @param col The color the honeycomb will be displayed
     */
    public void changeColor(java.awt.Color col)
    {
	cell.changeColor(col);
    }

    /**
     * You can select how polygons are drawn
     * @param t true = only Lines / false = real polygons
     */
    public void toLines(boolean t)
    {
	cell.toLines(t);
    }
    /**
     * returns if the honeycomb is marked
     * @return true if marked
     */
    public boolean isMarked()
    {
        return this.marked;
    }
    /**
     * A honeycomb can be marked by clicking it with the mouse and selected by
     * using the Mark and View Dialog
     * @param b marked or not
     * @param col the color the honeycomb gets marked with
     */
    public void setMarked(boolean b,java.awt.Color col)
    {
	if(b)
	{
            this.marked=true;
	    cell.markwithColor(col);
	    if(!col.equals(java.awt.Color.white));
	        this.selected=true;
	}
	else
	{
	    this.marked=false;
	    this.selected=false;
	    cell.demark();
	}
    }
    /**
     * @return If honeycomb is selected
     */
    public boolean isSelected()
    {
        return this.selected;
    }
    /**
     * Use this to (de)select a honeycomb
     * @param b selected yes or no
     */
    public void setSelected(boolean b)
    {
	this.selected=b;
	if(this.isMarked())return;
	if(selected)
	    cell.markwithColor(java.awt.Color.darkGray);
	else
	    cell.demark();
    }
    /**
     * This function is called when someone clicks on a honeycomb
     * @param event A mouse event
     */
    public void ClickActionAppeared(MouseEvent event)
    {
	if(event.getModifiers()==MouseEvent.BUTTON1_MASK)
	    viewport.reportMarkingOperationToDialogs(this.getpos());
	else if(event.getModifiers()==MouseEvent.BUTTON3_MASK)
	    viewport.translateToCoordinates(this.getpos());
    }

}