import java.io.*;
import java.util.*;

/**
 *
 * Stand: 09.07.2001
 */
public class Nest{    
    /**
     * Reprsentation eines leeren Blocks
     */
    public final static char LEER  = '.';

    /**
     * Reprsentation eines bebauten Blocks
     */
    public final static char STEIN = 'X';

    /**
     * Reprsentation eines Agenten
     */
    public final static char AGENT = 'a';

    /**
     * Reprsentation einer Wildcard
     */
    public final static char JOKER = 'j';

    /**
     * Reprsentation eines Blocks ausserhalb der Nest
     */
    public final static char OUT   = 'Z';


    /**
     * Projektweite '\n'-Konstanten, systemabhngig
     */
    public final static String NEWLINE = System.getProperty( "line.separator" );
    public final static String HEADER = "<HEAD>\377GENOM\377\260</HEAD>\377";

    // am 18.06. auf 3D umgestellt!
    protected Koord3D[] posAgent;
    protected Koord3D sichtweite;
    protected Koord3D posStart;           // hier wurde der erste Stein platziert. Sollte nur einmal gesetzt und nicht mehr verndert werden!

    protected Agent[] agent;
    protected Genom genom;

    protected int viewSize;

    protected AgentsWay[] agentsWays;

    protected char[][][] umgebung;

    protected int numberAgents;

    // added 12.07.2001
    public void finalize(){
        posAgent = null;
        sichtweite = null;
        posStart = null;
        agent = null;
        genom = null;
        agentsWays = null;
        umgebung = null;

        System.gc();
        System.gc();
    }


    protected int[] genCounter; // 13.07.2001

    /**
     * Konstruktor, der x- und y-Ausdehnung der Nest und das Genom fr den/die Agenten in dieser Nest erwartet.
     *
     * Gendert: 18.06.2001 z-Achse hinzugefgt
     *           12.07.2001 viewSize wurde mit aufgenommen
     *           13.07.2001 genCounter eingefhrt
     *           17.07.2001 die Art des CrystallizationPoints wurde bercksichtigt
     *
     * @param xsize Ausdehnung in x-Achse
     * @param ysize Ausdehnung in y-Achse
     * @param zsize Ausdehnung in z-Achse
     * @param genom Genom, das in den Agenten gesteckt werden soll
     * @param numbAgents die Anzahl der Agent, die in das Nest gesetzt werden sollen
     * @param viewSize die Grsse des Blickfelds der Agenten
     * @param crystallizationPoint die Art des Startpunkts: im Zentrum, in der Mitte (Decke), in der Ecke(0,0,0)
     */
    public Nest(int xsize, int ysize, int zsize, Genom genom, int numbAgents, int viewSize, int crystallizationPoint )
    {
	switch( crystallizationPoint ){
	case 0:
	    posStart = new Koord3D(xsize / 2,ysize / 2, zsize / 2);
	    break;
	case 1:
	    posStart = new Koord3D( xsize / 2, ysize / 2, 0 );
	    break;
	case 2:
	    posStart = new Koord3D( 0,0,0 );
	}
	   
        this.viewSize = viewSize;

        this.genom = genom;

        resetNest(xsize, ysize, zsize);
        setNumberAgents( numbAgents );

        agentsWays = new AgentsWay[numberAgents];

        initNest( genom, numbAgents, true );
    }

    /**
     * Dieser Konstruktor erwartet nur das Genom, das in den Agenten gesteckt werden soll. 
     *
     * Gendert: 18.06.2001 auf 3D umgestellt mit Standardwert
     *           02.07.2001 Multiagenten
     *           12.07.2001 viewSize wurde mit aufgenommen
     * @param genom Genom, das in den Agenten gesteckt werden soll
     * @param size Grsse des Nests
     * @param numbAgents die Anzahl der Agenten, die in die Nest sollen
     * @param viewSize die Sichtweite
     * @param crystallizationPoint Art des Startpunkts: im Zentrum, in der "Deckenmitte" oder in der Ecke (0,0,0)
     */
    public Nest( Genom genom, Koord3D size, int numbAgents, int viewSize, int crystallizationPoint )
    {
        this(size.getX(),size.getY(),size.getZ(), genom, numbAgents, viewSize, crystallizationPoint);
    }


    protected void setNumberAgents(int number ){
        if( number <= 0 ){
            message( "Fehler! Zu wenige Agenten in Nest!");
        }else{
            numberAgents = number;
        }
    }

    // 09.07.2001
    public int getNumberAgents(){
        return agent.length;
    }

    /**
     * Lscht die Neststruktur. Das verwendete Genom bleibt jedoch erhalten.
     *
     * Gendert: 21.06.2001: initNest() -> reInitNest()
     */
    public void reset(){
        resetNest( umgebung[0][0].length, umgebung[0].length, umgebung.length );
        reInitNest();
    }

    // posStart wird jetzt im Konstruktor initialisiert
    protected void resetNest( int xsize, int ysize, int zsize ){
        genCounter = new int[genom.size()]; 

        umgebung = new char[zsize][ysize][xsize];        
        for (int x = 0; x < xsize; x++){
            for (int y = 0; y < ysize; y++){
                for( int z = 0; z < zsize; z++){
                    umgebung[z][y][x] = LEER;
                }
            }
        }
        umgebung[posStart.getZ()][posStart.getY()][posStart.getX()] = STEIN;
    }

    protected void initNest(Genom genom, int numbAgents, boolean shallInitWays){
        posAgent = new Koord3D[numbAgents];
        agent = new Agent[numbAgents];  // Eingefgt 2.7. (Steffen)
        numberAgents = numbAgents;      // 03.07.2001


        if( shallInitWays )
            initAgentsWays();

        sichtweite = new Koord3D(viewSize,viewSize,viewSize); //Sichtweite der Agent sei 3x3x3
        Blickfeld view = new Blickfeld(new Koord3D((viewSize-1)/2,(viewSize-1)/2,(viewSize-1)/2), new Koord3D(viewSize,viewSize,viewSize), this );

        Knoten tree = new PatternGenerator( genom ).getTree();


        for( int i = 0; i<numbAgents; i++ ){
            createAndPositionNewAgent( i, posStart, view, tree );
            agentsWays[i].add( new Waypoint(posStart, true ));

            Koord3D firstMove = null;
            try{
                firstMove = agent[i].moveYou();

                agentsWays[i].add( new Waypoint( firstMove, false ));

                setPosAgent( i, firstMove );   // erstmal die Agent vom ersten Block wegbewegen. Sie bewegt sich auf jeden Fall, da sie ja auf einem Block sitzt!
            }catch( NoMoveLeftException e ){
                message( "Agent hat keine Zugmglichkeit mehr!" );
            }
        }
    }

    // wird von reset() aufgerufen. Damit wird der Pattern-Baum nicht mehr gelscht, muss also auch nicht mehr neu erstellt werden
    protected void reInitNest(){
        initAgentsWays();

        for( int i = 0; i<numberAgents; i++){
            agentsWays[i].add( new Waypoint( posStart, true )); // 11.06.: CrystallizationPoint wird gesetzt
            setRePosAgent( i, posStart );

            Koord3D firstMove = null;
            try{
                firstMove = agent[i].moveYou();

                agentsWays[i].add( new Waypoint( firstMove, false ));

                setPosAgent( i, firstMove );   // erstmal die Agent vom ersten Block wegbewegen. Sie bewegt sich auf jeden Fall, da sie ja auf einem Block sitzt!
            }catch( NoMoveLeftException e ){
                message( "Agent hat keine Zugmglichkeit mehr!" );
            }
        }
    }

    protected void initAgentsWays(){
        agentsWays = new AgentsWay[numberAgents];
        for( int i = 0; i<numberAgents;i++){
            agentsWays[i] = new AgentsWay();
        }
    }

    public Waypoint[] runOneRound(){
        Waypoint[] points = new Waypoint[numberAgents];

        for( int i =0; i<numberAgents; i++){
            try{
                points[i] = runOneRound( i );
            }catch(NoMoveLeftException e ){
            }
        }

        return points;
    }


    /**
     * Lsst den Agenten eine Runde laufen. Dies geschieht so: Der Agent sieht sich seine Nest an, dann bestimmt er,
     * ob er hier bauen soll oder nicht und setzt ggf. einen Stein. Die Nest merkt sich das (auch im agentsWay).
     * Danach wird der Agent bewegt.
     *
     * Gendert: 19.06.2001
     *           09.07.2001 einige Variablen eingespart (x,y,z, currentPos)
     *           13.07.2001 genCounter eingefhrt
     * @return WayPoint Der Punkt im Nest, den der Agent gerade besucht hat.
     */
    protected Waypoint runOneRound(int numberA)
        throws NoMoveLeftException{
        boolean built = false;

        agent[numberA].lookAround();

        int indexGen = agent[numberA].getGenUsed();
        if( indexGen >= 0 ){
            genCounter[indexGen]++;
        }

        built = agent[numberA].bOrNot2B();
        if( built ){
            umgebung[agent[numberA].getPosition().getZ()][agent[numberA].getPosition().getY()][agent[numberA].getPosition().getX()] = STEIN;
        }

        Waypoint currentRound = new Waypoint( agent[numberA].getPosition(), built );

        Koord3D newPosition =  agent[numberA].moveYou();
        setPosAgent( numberA, newPosition );

        agentsWays[numberA].add( currentRound );

        return currentRound;
    }

    // for debugging purpose -> Konstruktor, resetNest, runOneRound, Environment.laufeRunden
    /*
    protected String genCounterToString(){
    String result = new String();
    for( int i = 0; i<genCounter.length; i++ ){
    result += "Gen "+i+" wurde "+genCounter[i]+" mal angewandt.\n";
    }
    return result;
    }
    */

    public int getHowOftenUsed_Gen( int index ){
        if( genCounter != null ){
            return genCounter[index];
        }else{
            return -1;
        }
    }

    /**
     * Hiermit kann man die Konfliktauflsungsstrategie bestimmen, mit der gleiche/hnliche Pattern vorrang vor einander bekommen
     *
     * Gendert: 12.06.2001
     * @param strategy die gewnschte Strategie (0: Erste, 1: Zufllig, 2: Mit wenigsten Jokern, 3: Mit meisten Jokern)
     */
    public void setStrategy( int strategy ){
        for( int i = 0; i<numberAgents;i++){
            agent[i].setStrategy(strategy);
        }
    }

    /**
     * Vielleicht interessiert ja auch, welche Strategie gerade Verwendung findet.
     *
     * Gendert: 12.06.2001
     * @return gerade verwendete Strategie
     */
    public int getStrategy(){
        return agent[0].getStrategy();
    }

    /**
     * Hiermit kann die Position des Agenten verndert werden.
     *
     * Gendert: ?
     * @return AgentsWay den Weg, den der Agent bisher zurckgelegt hat. Darin enthalten ist die Information, wann er gebaut hat.
     */
    public AgentsWay getAgentsWay(int i){
        return agentsWays[i];
    }

    /**
     * Hiermit kann die Position des Agenten verndert werden.
     *
     * Gendert: ?
     * @param newPos Koord3D mit der neuen Position
     */
    protected void setPosAgent( int numberA, Koord3D newPos ){
        posAgent[numberA] = newPos;
    }

    // 28.06.2001 wird in reInitNest() aufgerufen
    protected void setRePosAgent( int numberA, Koord3D newPos ){
        posAgent[numberA] = newPos;
        agent[numberA].setPosition( newPos );
    }

    /**
     * Liefert die Art des Blocks an der angegebenen Position.
     *
     * Gendert: 18.06.2001 Rckgabewert jetzt char (Steffen)
     * @param where zu untersuchende Position
     * @return int Art des Blocks (siehe Konstanten oben)
     */
    public char getStateAtCoord( Koord3D where ){
        try {
            return umgebung[where.getZ()][ where.getY() ][ where.getX() ];
        } catch (ArrayIndexOutOfBoundsException e) {
            return OUT;
        }
    }

    protected void createAndPositionNewAgent(int numberA, Koord3D pos, Blickfeld view, Knoten tree)
    {
        setPosAgent( numberA, pos );
        agent[numberA] = new Agent(pos, sichtweite, view, tree);
    }

    /**
     * Speichert Informationen so, dass sie mit loadNestAndGenom rekonstruiert werden knnen.
     * (Da der Stock jetzt 3D ist, wird aus dem Stock nur ein Querschnitt visualisiert (zsize / 2).)
     *
     * Gendert: 18.06.2001 3D
     *           03.07.2001 Multiagenten
     * @param filename Dateiname (.txt)  samt Verzeichnis
     * @return ob das Speicher geklappt hat, oder nicht (IOException oder so)
     */
    public boolean saveNestAndGenom( String filename ){
        boolean noFailure = false;

        debug( "Nest.saveNestAndGenom() -> filename: "+filename );

        try{
            FileWriter writer = new FileWriter( filename, false );

            // Schreibe Header
            writer.write( HEADER + NEWLINE );

            // schreibe Genom
            String genom = this.genom.toString();
            writer.write( genom );
            writer.write( NEWLINE );

            writer.flush();
            writer.close();
            noFailure = true;
        }catch( IOException e ){

        }

        return noFailure;
    }

    /**
     * Liefert das Genom des Agenten
     *
     * Gendert: ?
     * @return verwendetes Genom
     */
    public Genom getGenom(){
        return genom;
    }

    public Koord3D getCrystallizationPoint(){
        return posStart;
    }

    // for debugging purpose
    public static void debug( String msg ){
        System.out.println( msg );
    }

    // Um Fehlermeldungen auszugeben
    public static void message( String msg ){
        System.err.println( msg );
    }
}
